
from typing import List, Tuple

def xwcwidth(c: str) -> int: ...
def cut_text(string: str, width: int) -> List[Tuple[int, int]]: ...
def wcswidth(string: str) -> int: ...
def cut_by_columns(string: str, limit: int) -> str: ...
